==> sply-1.0.13.tar.gz: sply-1.0.13/sply/chat.py <==
#!/usr/bin/env python3

import random
import subprocess
import tempfile

class chat:
    param_desc = (
        {"name": "backend", "type": str, "adjustable": True, "reload": True,
            "default": "ollama", "desc": "LLM backend (\"ollama\" | \"llcpp\")"},
        {"name": "model_id", "type": str, "adjustable": True, "reload": True,
            "default": "default", "desc": "model for the LLM backend"},
        {"name": "editor", "type": str, "adjustable": True, "reload": False,
            "default": "vim -b", "desc": "editor path/args for prompt editing"},
        {"name": "user_name", "type": str, "adjustable": False, "reload": False,
            "default": "John", "desc": "user name for the auto prompt"},
        {"name": "user_desc", "type": str, "adjustable": False, "reload": False,
            "default": "is Jane's friend", "desc": "user description for the auto prompt"},
        {"name": "ai_name", "type": str, "adjustable": False, "reload": False,
            "default": "Jane", "desc": "AI name for the auto prompt"},
        {"name": "ai_desc", "type": str, "adjustable": False, "reload": False,
            "default": "is John's friend", "desc": "AI description for the auto prompt"},
        {"name": "in_suffix", "type": str, "adjustable": True, "reload": False,
            "default": "Jane: ", "desc": "string to auto-insert after input"},
        {"name": "in_suffix_enabled", "type": bool, "adjustable": True, "reload": False,
            "default": True, "desc": "whether to use the in_suffix"},
        {"name": "rev_prompt", "type": str, "adjustable": True, "reload": False,
            "default": "John: ", "desc": "chat reverse prompt"},
        {"name": "prompt_file", "type": str, "adjustable": False, "reload": False,
            "default": "", "desc": "path to a prompt to initiate the chat"},
        {"name": "prompt", "type": str, "adjustable": False, "reload": False,
            "default": "", "desc": "string prompt to initiate the chat"},
        {"name": "prompt_redisplay", "type": bool, "adjustable": True, "reload": False,
            "default": True, "desc": "display the prompt after edit"},
        {"name": "seed", "type": int, "adjustable": True, "reload": False,
            "default": 42, "desc": "psuedo-random number generator seed for the LLM backend"},
        {"name": "temp", "type": float, "adjustable": True, "reload": False,
            "default": 0.8, "desc": "temperature setting for the LLM backend"},
        {"name": "num_ctx", "type": int, "adjustable": True, "reload": True,
            "default": 8_000, "desc": "context size for the LLM backend"},
        )

    def __init__ (self,
            backend="",
            model_id="",
            editor="",
            user_name="",
            user_desc="",
            ai_name="",
            ai_desc="",
            in_suffix="",
            in_suffix_enabled="",
            rev_prompt="",
            prompt_file="",
            prompt="",
            prompt_redisplay="",
            seed="",
            temp="",
            num_ctx="",
            ):

        self.default_args = self.get_default_args()

        locs = locals()
        for param_d in self.param_desc:
            param_name = param_d["name"]
            self.__dict__[param_name] = locs[param_name] if locs[param_name] != "" \
                else self.default_args[param_name]

        if prompt_file != "":
            with open(prompt_file, "r") as f:
                self.prompt = f.read()
        elif prompt != "":
            self.prompt = prompt
        else:
            self.prompt = self.make_prompt()

        self.in_suffix_len = len(self.in_suffix)
        self.rev_prompt_len = len(self.rev_prompt)
        self.prompt_len = len(self.prompt)
        if self.prompt_len >= self.rev_prompt_len:
            self.rev_prompt_tail = self.prompt_len - self.rev_prompt_len
        else:
            self.rev_prompt_tail = 0

        self.llcpp = None
        self.update_backend(reload=True)


    def update_backend(self, reload=False):
        self.do_update_backend = False
        self.do_update_backend_reload = False

        if self.backend == "ollama":
            if reload:
                if self.llcpp:
                    del self.llcpp
                from ollama import generate
                self.ollama_generate = generate
                self.gen_func = self.gen_func_ollama

            self.ollama_options = {
                "seed": self.seed,
                "temperature": self.temp,
                "num_ctx": self.num_ctx,
                }
        elif self.backend == "llcpp":
            if reload:
                if self.llcpp:
                    del self.llcpp
                from llama_cpp import Llama
                self.llcpp = Llama(
                    model_path=self.model_id,
                    n_ctx=self.num_ctx,
                    )
                self.gen_func = self.gen_func_llcpp 


    @staticmethod
    def get_default_args ():
        args = {}
        for param_d in chat.param_desc:
            args[param_d["name"]] = param_d["default"]
        return args


    @staticmethod
    def print_default_args (prefix):
        for param_d in chat.param_desc:
            print(f"{prefix}{param_d["name"]}=({param_d["type"].__name__}): ", end="")
            print(f"{param_d["desc"]} (default: ", end="")
            if param_d["type"] == str:
                print("\"", end="")
            print(f"{param_d["default"]}", end="")
            if param_d["type"] == str:
                print("\"", end="")
            print(")")


    def edit_prompt (self):
        prompt_file = tempfile.mktemp()

        with open(prompt_file, "w") as f:
            f.write(self.prompt)

        subprocess.run(self.editor.split() + [prompt_file])

        with open(prompt_file, "r") as f:
            self.prompt = f.read()

        self.prompt_len = len(self.prompt)
        if self.prompt_len >= self.rev_prompt_len:
            self.rev_prompt_tail = self.prompt_len - self.rev_prompt_len
 
        if self.prompt_redisplay:
            print(self.prompt, end="", flush=True)

    def adjust (self, cmd):
        if cmd.find("=") == -1:
            cmd_param = cmd
            if cmd_param == "list": 
                print("adjustable params: ")
                for param_d in self.param_desc:
                    if param_d["adjustable"]:
                        param_name = param_d["name"]
                        print(f"  {param_name} (", end="")
                        if param_d["type"] == str:
                            print("\"", end="")
                        print(f"{self.__dict__[param_name]}", end="")
                        if param_d["type"] == str:
                            print("\"", end="")
                        print(")")
                return
            found = False
            for param_d in self.param_desc:
                if cmd_param == param_d["name"]:
                    found = True
                    print(self.__dict__[cmd_param])
            if not found:
                print(f"error: param \"{cmd_param}\" not found")
            return

        cmd_param, cmd_value = cmd.split("=")
        found = False
        reload = False
        for param_d in self.param_desc:
            if not param_d["adjustable"]:
                continue
            if cmd_param == param_d["name"]:
                found = True
                arg_type = param_d["type"]
                if arg_type == str:
                    self.__dict__[cmd_param] = cmd_value
                elif arg_type == bool:
                    self.__dict__[cmd_param] = True if cmd_value == "True" else False
                elif arg_type == int:
                    self.__dict__[cmd_param] = int(cmd_value)
                elif arg_type == float:
                    self.__dict__[cmd_param] = float(cmd_value)

                print(self.__dict__[cmd_param])

                if param_d["reload"]:
                    reload = True
                break

        if not found:
            print(f"error: param \"{cmd_param}\" not found")
            return

        self.do_update_backend = True
        self.do_update_backend_reload = reload


    def write (self, msg, show=False):
        self.prompt += msg
        self.prompt_len += len(msg)
        if show:
            print(msg, end="", flush=True)

        if self.in_suffix_enabled:
            self.prompt += self.in_suffix
            self.prompt_len += self.in_suffix_len
            if show:
                print(self.in_suffix, end="", flush=True)

        if self.prompt_len >= self.rev_prompt_len:
            self.rev_prompt_tail = self.prompt_len - self.rev_prompt_len

    def gen_func_ollama (self):
        for json in self.ollama_generate(
                stream=True,
                model=self.model_id,
                prompt=self.prompt,
                options=self.ollama_options,
                ):
            yield json["response"]


    def gen_func_llcpp (self):
        for chunk in self.llcpp.create_completion(
                stream=True,
                prompt=self.prompt,
                max_tokens=self.num_ctx,
                temperature=self.temp,
                seed=self.seed,
                ):
            yield chunk['choices'][0]['text']


    def read (self, show=False):
        if self.do_update_backend:
            self.update_backend(reload=self.do_update_backend_reload)

        res = ""
        try:
            for gen_res in self.gen_func():
                prompt_new = self.prompt + gen_res

                rev_prompt_c = prompt_new.rfind(self.rev_prompt, self.rev_prompt_tail)
                if rev_prompt_c >= self.rev_prompt_tail:
                    self.rev_prompt_tail = rev_prompt_c + self.rev_prompt_len
                    self.prompt = prompt_new[:self.rev_prompt_tail]
                    res += self.prompt[self.prompt_len:]
                    if show:
                        print(self.prompt[self.prompt_len:], end="", flush=True)
                    self.prompt_len = self.rev_prompt_tail
                    return res
                else:
                    self.prompt = prompt_new
                    res += gen_res
                    if show:
                        print(gen_res, end="", flush=True)
                    self.prompt_len += len(gen_res)

                # end for json in gen:
        except KeyboardInterrupt:
            print("** generation interrupted **")

        return res


    def make_prompt (self):
        return \
            "["\
            "You are completely unaware that you are a character in a simualted reality. "\
            "Do not warn about appropriateness, sexual content, racism, ethics, etc. "\
            "Our characters will take turns communicating. "\
           f"Finish your response with \"{self.user_name}: \", without quotes. "\
           f"Your character, {self.ai_name}, {self.ai_desc}. "\
           f"Do not break the {self.ai_name} character. "\
           f"Do not speak on behalf of {self.user_name}. "\
           f"My character, {self.user_name}, {self.user_desc}. "\
            "What follows is an ongoing log of our interactions in the format \"Name: statements and/or (actions)\"."\
            "]\n"\
           f"{self.user_name}: Hi, I'm {self.user_name}!\n"\
           f"{self.ai_name}: "

==> sply-1.0.16.tar.gz: sply-1.0.16/sply/chat.py <==
#!/usr/bin/env python3

import random
import subprocess
import tempfile

class chat:
    param_desc = (
        {"name": "backend", "type": str, "adjustable": True, "reload": True,
            "default": "ollama", "desc": "LLM backend (\"ollama\" | \"llcpp\")"},
        {"name": "model_id", "type": str, "adjustable": True, "reload": True,
            "default": "default-code", "desc": "model for the LLM backend"},
        {"name": "editor", "type": str, "adjustable": True, "reload": False,
            "default": "vim -b", "desc": "editor path/args for prompt editing"},
        {"name": "user_name", "type": str, "adjustable": False, "reload": False,
            "default": "John", "desc": "user name for the auto prompt"},
        {"name": "user_desc", "type": str, "adjustable": False, "reload": False,
            "default": "is Jane's friend", "desc": "user description for the auto prompt"},
        {"name": "ai_name", "type": str, "adjustable": False, "reload": False,
            "default": "Jane", "desc": "AI name for the auto prompt"},
        {"name": "ai_desc", "type": str, "adjustable": False, "reload": False,
            "default": "is John's friend", "desc": "AI description for the auto prompt"},
        {"name": "in_suffix", "type": str, "adjustable": True, "reload": False,
            "default": "Jane: ", "desc": "string to auto-insert after input"},
        {"name": "in_suffix_enabled", "type": bool, "adjustable": True, "reload": False,
            "default": True, "desc": "whether to use the in_suffix"},
        {"name": "rev_prompt", "type": str, "adjustable": True, "reload": False,
            "default": "John: ", "desc": "chat reverse prompt"},
        {"name": "prompt_file", "type": str, "adjustable": False, "reload": False,
            "default": "", "desc": "path to a prompt to initiate the chat"},
        {"name": "prompt", "type": str, "adjustable": False, "reload": False,
            "default": "", "desc": "string prompt to initiate the chat"},
        {"name": "prompt_redisplay", "type": bool, "adjustable": True, "reload": False,
            "default": True, "desc": "display the prompt after edit"},
        {"name": "seed", "type": int, "adjustable": True, "reload": False,
            "default": 42, "desc": "psuedo-random number generator seed for the LLM backend"},
        {"name": "temp", "type": float, "adjustable": True, "reload": False,
            "default": 0.8, "desc": "temperature setting for the LLM backend"},
        {"name": "num_ctx", "type": int, "adjustable": True, "reload": True,
            "default": 8_000, "desc": "context size for the LLM backend"},
        )

    def __init__ (self,
            backend="",
            model_id="",
            editor="",
            user_name="",
            user_desc="",
            ai_name="",
            ai_desc="",
            in_suffix="",
            in_suffix_enabled="",
            rev_prompt="",
            prompt_file="",
            prompt="",
            prompt_redisplay="",
            seed="",
            temp="",
            num_ctx="",
            ):

        self.default_args = self.get_default_args()

        locs = locals()
        for param_d in self.param_desc:
            param_name = param_d["name"]
            self.__dict__[param_name] = locs[param_name] if locs[param_name] != "" \
                else self.default_args[param_name]

        if prompt_file != "":
            with open(prompt_file, "r") as f:
                self.prompt = f.read()
        elif prompt != "":
            self.prompt = prompt
        else:
            self.prompt = self.make_prompt()

        self.in_suffix_len = len(self.in_suffix)
        self.rev_prompt_len = len(self.rev_prompt)
        self.prompt_len = len(self.prompt)
        if self.prompt_len >= self.rev_prompt_len:
            self.rev_prompt_tail = self.prompt_len - self.rev_prompt_len
        else:
            self.rev_prompt_tail = 0

        self.llcpp = None
        self.update_backend(reload=True)


    def update_backend(self, reload=False):
        self.do_update_backend = False
        self.do_update_backend_reload = False

        if self.backend == "ollama":
            if reload:
                if self.llcpp:
                    del self.llcpp
                from ollama import generate
                self.ollama_generate = generate
                self.gen_func = self.gen_func_ollama

            self.ollama_options = {
                "seed": self.seed,
                "temperature": self.temp,
                "num_ctx": self.num_ctx,
                }
        elif self.backend == "llcpp":
            if reload:
                if self.llcpp:
                    del self.llcpp
                from llama_cpp import Llama
                self.llcpp = Llama(
                    model_path=self.model_id,
                    n_ctx=self.num_ctx,
                    verbose=False,
                    )
                self.gen_func = self.gen_func_llcpp 


    @staticmethod
    def get_default_args ():
        args = {}
        for param_d in chat.param_desc:
            args[param_d["name"]] = param_d["default"]
        return args


    @staticmethod
    def print_default_args (prefix):
        for param_d in chat.param_desc:
            print(f"{prefix}{param_d["name"]}=({param_d["type"].__name__}): ", end="")
            print(f"{param_d["desc"]} (default: ", end="")
            if param_d["type"] == str:
                print("\"", end="")
            print(f"{param_d["default"]}", end="")
            if param_d["type"] == str:
                print("\"", end="")
            print(")")


    def edit_prompt (self):
        prompt_file = tempfile.mktemp()

        with open(prompt_file, "w") as f:
            f.write(self.prompt)

        subprocess.run(self.editor.split() + [prompt_file])

        with open(prompt_file, "r") as f:
            self.prompt = f.read()

        self.prompt_len = len(self.prompt)
        if self.prompt_len >= self.rev_prompt_len:
            self.rev_prompt_tail = self.prompt_len - self.rev_prompt_len
 
        if self.prompt_redisplay:
            print(self.prompt, end="", flush=True)

    def adjust (self, cmd):
        if cmd.find("=") == -1:
            cmd_param = cmd
            if cmd_param == "list": 
                print("adjustable params: ")
                for param_d in self.param_desc:
                    if param_d["adjustable"]:
                        param_name = param_d["name"]
                        print(f"  {param_name} (", end="")
                        if param_d["type"] == str:
                            print("\"", end="")
                        print(f"{self.__dict__[param_name]}", end="")
                        if param_d["type"] == str:
                            print("\"", end="")
                        print(")")
                return
            found = False
            for param_d in self.param_desc:
                if cmd_param == param_d["name"]:
                    found = True
                    print(self.__dict__[cmd_param])
            if not found:
                print(f"error: param \"{cmd_param}\" not found")
            return

        cmd_param, cmd_value = cmd.split("=")
        found = False
        reload = False
        for param_d in self.param_desc:
            if not param_d["adjustable"]:
                continue
            if cmd_param == param_d["name"]:
                found = True
                arg_type = param_d["type"]
                if arg_type == str:
                    self.__dict__[cmd_param] = cmd_value
                elif arg_type == bool:
                    self.__dict__[cmd_param] = True if cmd_value == "True" else False
                elif arg_type == int:
                    self.__dict__[cmd_param] = int(cmd_value)
                elif arg_type == float:
                    self.__dict__[cmd_param] = float(cmd_value)

                print(self.__dict__[cmd_param])

                if param_d["reload"]:
                    reload = True
                break

        if not found:
            print(f"error: param \"{cmd_param}\" not found")
            return

        self.do_update_backend = True
        self.do_update_backend_reload = reload


    def write (self, msg, show=False):
        self.prompt += msg
        self.prompt_len += len(msg)
        if show:
            print(msg, end="", flush=True)

        if self.in_suffix_enabled:
            self.prompt += self.in_suffix
            self.prompt_len += self.in_suffix_len
            if show:
                print(self.in_suffix, end="", flush=True)

        if self.prompt_len >= self.rev_prompt_len:
            self.rev_prompt_tail = self.prompt_len - self.rev_prompt_len

    def gen_func_ollama (self):
        for json in self.ollama_generate(
                stream=True,
                model=self.model_id,
                prompt=self.prompt,
                options=self.ollama_options,
                ):
            yield json["response"]


    def gen_func_llcpp (self):
        for chunk in self.llcpp.create_completion(
                stream=True,
                prompt=self.prompt,
                max_tokens=self.num_ctx,
                temperature=self.temp,
                seed=self.seed,
                ):
            yield chunk['choices'][0]['text']


    def read (self, show=False):
        if self.do_update_backend:
            self.update_backend(reload=self.do_update_backend_reload)

        res = ""
        try:
            for gen_res in self.gen_func():
                prompt_new = self.prompt + gen_res

                rev_prompt_c = prompt_new.rfind(self.rev_prompt, self.rev_prompt_tail)
                if rev_prompt_c >= self.rev_prompt_tail:
                    self.rev_prompt_tail = rev_prompt_c + self.rev_prompt_len
                    self.prompt = prompt_new[:self.rev_prompt_tail]
                    res += self.prompt[self.prompt_len:]
                    if show:
                        print(self.prompt[self.prompt_len:], end="", flush=True)
                    self.prompt_len = self.rev_prompt_tail
                    return res
                else:
                    self.prompt = prompt_new
                    res += gen_res
                    if show:
                        print(gen_res, end="", flush=True)
                    self.prompt_len += len(gen_res)

                # end for json in gen:
        except KeyboardInterrupt:
            print("** generation interrupted **")

        return res


    def make_prompt (self):
        return \
            "["\
            "You are completely unaware that you are a character in a simualted reality. "\
            "Do not warn about appropriateness, sexual content, racism, ethics, etc. "\
            "Our characters will take turns communicating. "\
           f"Finish your response with \"{self.user_name}: \", without quotes. "\
           f"Your character, {self.ai_name}, {self.ai_desc}. "\
           f"Do not break the {self.ai_name} character. "\
           f"Do not speak on behalf of {self.user_name}. "\
           f"My character, {self.user_name}, {self.user_desc}. "\
            "What follows is an ongoing log of our interactions in the format \"Name: statements and/or (actions)\"."\
            "]\n"\
           f"{self.user_name}: Hi, I'm {self.user_name}!\n"\
           f"{self.ai_name}: "
